#pragma once
#include "ikfksolver.h"

class FKCtrl : public Control, public IKFKSolverInterface
{
	public:


		// Parameter block
		IParamBlock2	*prefPblock;
		Control			*FK;
		IParamBlock2	*pblock;
		IKFKSolver		*solver;
		Matrix3 value;
		Interval ivalid;

		IObjParam *ip;
		HWND hwnd;
		BOOL mustdelete;
		BOOL sendingMessage;
		BOOL allKeysActive;

		//Animatable Methods
		Class_ID ClassID() {return FKCtrl_CLASS_ID;}		
		SClass_ID SuperClassID() { return CTRL_MATRIX3_CLASS_ID; }
		void GetClassName(TSTR& s) {s = GetString(IDS_FK_CLASS_NAME);}
		int NumSubs() {return CTRL_NSUBS;}
		Animatable* SubAnim(int i);
		TSTR SubAnimName(int i);	
		int SubNumToRefNum(int subNum) {  if (subNum==LOCALCTRL_REF) return LOCALCTRL_REF; else return -1;}
		BOOL AssignController(Animatable *control,int subAnim);
		BOOL IsReplaceable  () {return FALSE;}
		BOOL CanCopyAnim  () {return FALSE;}
		BOOL CanInstanceController  () {return FALSE;}

		//to display controller params on the motion panel 
		void BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev ); 
		// to undisplay 
		void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next ); 

		// return number of ParamBlocks in this instance 
		int NumParamBlocks() {return 1;}
		IParamBlock2* GetParamBlock(int i) {if (i==0) return prefPblock; else return	NULL;} 
		// return id'd ParamBlock 
		IParamBlock2* GetParamBlockByID(BlockID id) {
			if (prefPblock && prefPblock->ID()==id)
				return prefPblock;
			return NULL;
		}

		//Reference Methods
		int NumRefs() {return FK_REFS;}
		RefTargetHandle GetReference(int i);
		void SetReference(int i, RefTargetHandle rtarg);
#if IS_SUPMAX2015
		RefResult NotifyRefChanged(const Interval& changeInt, RefTargetHandle hTarget, 
			PartID& partID,  RefMessage message, BOOL propagate);
#else
		RefResult NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
			PartID& partID,  RefMessage message);
#endif
		IOResult Load(ILoad *iload) {return IO_OK;}
		IOResult Save(ISave *isave) {return IO_OK;}

		//Control Methods
		void Copy(Control *from);
		RefTargetHandle Clone( RemapDir &remap );
		int IsKeyable(){return TRUE;}
		BOOL IsLeaf(){return FALSE;}
		void GetValue(TimeValue t, void *val, Interval &valid, GetSetMethod method = CTRL_ABSOLUTE );
		void SetValue(TimeValue t, void *val, int commit=1, GetSetMethod method= CTRL_ABSOLUTE);
		void CommitValue(TimeValue t);
		void RestoreValue(TimeValue t);
		BOOL ChangeParents  (TimeValue  t,const Matrix3 &oldP, const Matrix3 &newP, const Matrix3 &tm);
		Control* GetPositionController() {if (FK) return FK->GetPositionController(); else return NULL;}
		Control* GetRotationController() {if (FK) return FK->GetRotationController(); else return NULL;}
		Control* GetScaleController() {if (FK) return FK->GetScaleController(); else return NULL;}
		BOOL SetPositionController(Control *c){if (FK) return FK->SetPositionController(c); else return FALSE;}
		BOOL SetRotationController(Control *c){if (FK) return FK->SetRotationController(c); else return FALSE;}
		BOOL SetScaleController(Control *c){if (FK) return FK->SetScaleController(c); else return FALSE;}
		


		//keys functions
		int  NumKeys (){if (solver && allKeysActive) return solver->NumKeys(); else return 0;}
		TimeValue  GetKeyTime (int index){if (solver && allKeysActive) return solver->GetKeyTime(index); else return 0;}
		int  GetKeyIndex (TimeValue t){if (solver && allKeysActive) return solver->GetKeyIndex(t); else return -1;}
		BOOL  GetNextKeyTime (TimeValue t, DWORD flags, TimeValue &nt) {if (solver && allKeysActive) return solver->GetNextKeyTime(t,flags,nt); else return FALSE;}
		void  CopyKeysFromTime (TimeValue src, TimeValue dst, DWORD flags) {if (solver && allKeysActive) solver->CopyKeysFromTime(src,dst,flags);}
		void  DeleteKeyAtTime (TimeValue t) {if (solver && allKeysActive) solver->DeleteKeyAtTime(t);}
		BOOL  IsKeyAtTime (TimeValue t, DWORD flags) {if (solver && allKeysActive) return solver->IsKeyAtTime(t,flags); else return FALSE;}
		int  GetKeyTimes (Tab< TimeValue > &times, Interval range, DWORD flags) {if (solver && allKeysActive) return solver->GetKeyTimes(times,range,flags); else return 0;}
		int  GetKeySelState (BitArray &sel, Interval range, DWORD flags) {if (solver && allKeysActive) return solver->GetKeySelState(sel,range,flags); else return 0;}
		void  MapKeys (TimeMap *map, DWORD flags) {if (solver && allKeysActive) solver->MapKeys(map,flags);}
		void  DeleteKeys (DWORD flags) {if (solver && allKeysActive) solver->DeleteKeys(flags);}
		void  DeleteKeyByIndex (int index) {if (solver && allKeysActive) solver->DeleteKeyByIndex(index);}
		void  SelectKeys (TrackHitTab &sel, DWORD flags) {if (solver && allKeysActive) solver->SelectKeys(sel,flags);}
		void  SelectKeyByIndex (int i, BOOL sel) {if (solver && allKeysActive) solver->SelectKeyByIndex(i,sel);}
		BOOL  IsKeySelected (int i) {if (solver && allKeysActive) return solver->IsKeySelected(i); else return FALSE;}
		void  FlagKey (TrackHitRecord hit){if (solver && allKeysActive) solver->FlagKey(hit);}
		int  GetFlagKeyIndex () {if (solver && allKeysActive) return solver->GetFlagKeyIndex(); else return -1;}
		int  NumSelKeys () {if (solver && allKeysActive) return solver->NumSelKeys(); else return 0;}
		void  CloneSelectedKeys (BOOL offset=FALSE) {if (solver && allKeysActive) solver->CloneSelectedKeys(offset);}
		void  AddNewKey (TimeValue t, DWORD flags){if (solver && allKeysActive) solver->AddNewKey(t,flags);}


		//published fns
		void Match();
		void SetFK();
		void SetIK();
		INode *GetMainFKNode(){if (solver) return solver->GetMainNode(); else return NULL;}
		INode *GetSecFKNode(){if (solver) return solver->GetSecNode(); else return NULL;}
		INode *GetLastFKNode(){if (solver) return solver->GetLastNode(); else return NULL;}
		INode *GetGoal(){if (solver) return solver->GetGoalNode(); else return NULL;}
		INode *GetTarget(){if (solver) return solver->GetTargetNode(); else return NULL;}
		INode *GetGoalParent(){if (solver && solver->GoalParentCount()) return solver->GetGoalParentNode(0); else return NULL;}
		void SetTargetNode(INode *theNode); 
		void SetGoalParentNode(INode *theNode); 

		FPInterfaceDesc* GetDesc();

		//FPInterface methods
		FPInterface* GetInterface(Interface_ID id) { 
		 if (id == IKFKSOLVER_INTERFACE) 
		   return (IKFKSolverInterface*)this; 
		 else 
			 return (FPInterface*)Control::GetInterface(id); 
		 } 

		//FKCtrl methods
		void SetInternalValue(Matrix3 thevalue,Interval theivalid){value=thevalue;ivalid=theivalid;}
		void NotifyChange(){NotifyDependents(FOREVER,PART_ALL,REFMSG_CHANGE);}
		

		//Constructor/Destructor
		FKCtrl(BOOL loading=FALSE);
		FKCtrl(IKFKSolver *theSolver);
		~FKCtrl();
};

class FKCtrlClassDesc : public ClassDesc2 
{

public:
	virtual int IsPublic() 							{ return FALSE; }
	virtual void* Create(BOOL loading) 		{return new FKCtrl(loading); }
	void DeleteThis() { delete this; }
	virtual const TCHAR *	ClassName() 			{ return GetString(IDS_FK_CLASS_NAME); }
	virtual SClass_ID SuperClassID() 				{ return CTRL_MATRIX3_CLASS_ID; }
	virtual Class_ID ClassID() 						{ return FKCtrl_CLASS_ID; }
	virtual const TCHAR* Category() 				{ return _T(""); }

	virtual const TCHAR* InternalName() 			{ return _T("FKCtrl"); }	// returns fixed parsable name (scripter-visible name)
	virtual HINSTANCE HInstance() 					{ return hInstance; }					// returns owning module handle
};